#define std_io_FileDescriptor_IMPORT
#define std_io_InputTube_IMPORT
#define std_io_OutputTube_IMPORT
#define std_lang_Object_IMPORT

#include <stdlib.h>
#include <unistd.h>
#include <std/io/impl/FileDescriptor.h>
#include <std/io/impl/InputTube.h>
#include <std/io/impl/OutputTube.h>

struct FileDescriptor_Closing {
    int (*close)(int);
};

static int
dontClose(int fd UNUSED)
{
    return 0;
}

static struct FileDescriptor_Closing noneClosing = {dontClose};
static struct FileDescriptor_Closing autoClosing = {close};
const struct FileDescriptor_Closing *const FileDescriptor_AUTO = &autoClosing;
const struct FileDescriptor_Closing *const FileDescriptor_NONE = &noneClosing;

struct impl_FDInputTube {
    union {
        struct impl_Object object;
        struct impl_InputTube inputTube;
    } super;
    int fd;
    const struct FileDescriptor_Closing *closing;
};

union FDIT {
    struct impl_InputTube *inputTube;
    struct impl_FDInputTube *impl;
} __attribute__((__transparent_union__));

static ssize_t
FDIT_read(union FDIT this, void *data, size_t size)
{
    struct impl_FDInputTube *impl = this.impl;
    int fd = impl->fd;

    return read(fd, data, size);
}

static int
FDIT_close(union FDIT this)
{
    struct impl_FDInputTube *impl = this.impl;
    int s;

    s = impl->closing->close(impl->fd);
    free(impl);
    return s;
}

static const struct InputTube_Methods FDIT_METHODS = {
    .read = FDIT_read,
    .close = FDIT_close};

union InputTube
FileDescriptor_createInputTube(
    int fd, const struct FileDescriptor_Closing *const closing)
{
    struct impl_FDInputTube *impl;

    if ((impl = malloc(sizeof(*impl))) == NULL) {
	return (union InputTube)Object_NULL;
    }
    InputTube_initialize(&impl->super.inputTube, &FDIT_METHODS);
    impl->fd = fd;
    impl->closing = closing;
    return (union InputTube)&impl->super.inputTube;
}

struct impl_FDOutputTube {
    union {
        struct impl_Object object;
        struct impl_OutputTube outputTube;
    } super;
    int fd;
    const struct FileDescriptor_Closing *closing;
};

union FDOT {
    struct impl_OutputTube *outputTube;
    struct impl_FDOutputTube *impl;
} __attribute__((__transparent_union__));

static ssize_t
FDOT_write(union FDOT this, const void *data, size_t size)
{
    struct impl_FDOutputTube *impl = this.impl;
    int fd = impl->fd;

    return write(fd, data, size);
}

static int
FDOT_close(union FDOT this)
{
    struct impl_FDOutputTube *impl = this.impl;
    int s;

    s = impl->closing->close(impl->fd);
    free(impl);
    return s;
}

static const struct OutputTube_Methods FDOT_METHODS = {
    .write = FDOT_write,
    .close = FDOT_close};

union OutputTube
FileDescriptor_createOutputTube(
    int fd, const struct FileDescriptor_Closing *const closing)
{
    struct impl_FDOutputTube *impl;

    if ((impl = malloc(sizeof(*impl))) == NULL) {
        return (union OutputTube)Object_NULL;
    }
    OutputTube_initialize(&impl->super.outputTube, &FDOT_METHODS);
    impl->fd = fd;
    impl->closing = closing;
    return (union OutputTube)&impl->super.outputTube;
}
